// (c) 2005 Witold Bot :: wersja DU
//	C99 version ...
// www.hope.art.pl

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

/*	TYPY DANYCH		*/

typedef struct Node *Pnode;
typedef struct Node {
    uint32_t freq;
    uint8_t c;
    uint8_t *code;
    Pnode left;
    Pnode right;
} TNode;

typedef struct Qrecord *Precord;
typedef struct Qrecord {
	Pnode   node;
	Precord next;
	Precord prev;
} TQrecord;


/*	ZMIENNE GLOBALE		*/

Precord Queue=NULL;		// kolejka - inicjowana przez Qinit()
Pnode Tree=NULL;		// drzewo - inicjowane przez HuffmanMakeTree()

uint32_t HuffmanCompressedLen = 0;	// dlugosc (w bitach) skompresowanego pliku
uint32_t RealLen = 0;		// dlugosc (w bitach) nieskompresowanego pliku


/* 			QUEUE			*/

// tworzymy wartownika kolejki - tzn. pusty rekord ktory oznacza
// poczatek kolejki - w ten sposob zawsze mozemy dosatawiac do
// kolejki ZA tym rekordem - przez co wstawianie staje sie latwe!
void Qinit() {
	Queue=(Precord)calloc(1,sizeof(TQrecord));
	Queue->node = calloc(1,sizeof(TNode));
}



// funkcja wstawia do kolejki gotowy node
// dzieki temu ze mamy wartownika wiemy, ze
// 1. temp nigdy nie jest NULL
// 2. zawsze mozemy wstawiac za tempem nawet jesli kolejka jest ,,pusta''
// 3. wstawiamy wiec ZA tempem ;)
void Qinsert2(Pnode node)
{
	Precord new;

// dzieki Qinit() temp nigdy nie jest NULL!
	Precord temp=Queue;

// tworzymy nowy rekord
	new=(Precord)calloc(1,sizeof(TQrecord));
	new->node=node;

// szukamy miejsca zeby wstawic
// temp bedzie tym rekokred ZA ktorym bedzie stac nasz nowy!
// dzieki Qinit() zawsze jest taka mozliwosc
	while((temp->next!=NULL)&&(temp->next->node!=NULL)&&(temp->next->node->freq < node->freq)) 
			temp=temp->next;

// naprawiamy polaczenia
	new->prev = temp;
	new->next = temp->next;
	new->prev->next = new;
	if(new->next!=NULL)new->next->prev = new;

}

// funkcja wstawia do kolejki pare freq:c
// jej dzialanie ogranicza sie do stworzenia nowego node'a i wykonaniu Qinsert2()
void Qinsert(uint32_t freq, uint8_t c)
{
	Pnode new;
	new=(Pnode)calloc(1,sizeof(TNode));
	new->freq=freq;
	new->c=c;
	Qinsert2(new);
}


// zwracamay ju Pnode - czyli kawaek drzewa ;)
// przy okazji oszczdzamy troch pamici ;)
Pnode QextractMin() {
        Precord Rtemp=Queue->next;
        Pnode Ntemp=Rtemp->node;

        if(Queue->next==NULL) {	// to nie moze sie zdarzyc! :)
        	printf("ERROR: proba zdjecia z pustej kolejki!\n");
        	exit(-1);
        }

        Queue->next=Queue->next->next;
        if(Queue->next!=NULL)Queue->next->prev=Queue;
        free(Rtemp);
        return Ntemp;
}

/*		HUFFMAN		*/
// zliczamy liczbe wystapien	
uint32_t* HuffmanCount(uint8_t *filename) {
	FILE *infile = fopen(filename,"rb");	// otwieramy plik binarnie
						// zeby nie bylo klopotow z
						// dziwnymi znaczkami ...
	uint32_t *znaki=(uint32_t*)calloc(256,sizeof(uint32_t));
	uint8_t znak;
	
	if (infile == NULL) {
		printf("ERROR: loading file %s failed!\n",filename);
		exit(-1);
	}


	while((!feof(infile))&&(fread(&znak,sizeof(uint8_t),1,infile))) {
		znaki[(int)znak]++;
	}

	fclose(infile);
	return(znaki);
}

// gdy drzewo jest gotowe - wyliczamy kody kazdego z wierzcholkow 
// rekurencyjnie poczawszy od korzenia ...
// rekurencja jest prosta - jesli wierzcholek ma kod == C
// to lewy syn musi miec kod = C0 a prawy C1
// ta funkcja odwiedzamy cale drzewo - w momencie napotkania
// liscie wypisujemy na ekran informacje o kodzie...
void HuffmanMakeCodes(Pnode root) {
	if(root->left!=NULL) {
		root->left->code = (uint8_t*)calloc(strlen(root->code)+2,sizeof(uint8_t));
		sprintf(root->left->code,"%s0",root->code); 
		HuffmanMakeCodes(root->left);
	}

	if(root->right!=NULL) {
		root->right->code = (uint8_t*)calloc(strlen(root->code)+2,sizeof(uint8_t));
		sprintf(root->right->code,"%s1",root->code); 
		HuffmanMakeCodes(root->right);
	}

	if(root->left==NULL && root->right==NULL) {
#ifdef STATS
	printf("char: (%d)\tfreq: %d\t\tcode: %s\n",root->c,root->freq,root->code);
#endif
	HuffmanCompressedLen+=root->freq * strlen(root->code);
	RealLen+=root->freq * 8;
	}
}

// funcja buduje drzewo - ale nie liczy kodow! wzywana po to jest
// funkcja HuffmanMakeCodes()
void HuffmanMakeTree(int charsCount) {
	Pnode left,right;
	Pnode temp;

	for(int i=0;i<charsCount-1;i++) {
		left=QextractMin();
		right=QextractMin();
		temp = (Pnode)calloc(1,sizeof(TNode));
		temp->left=left;
		temp->right=right;
		temp->freq = left->freq + right->freq;
//		temp->c = '!';	// tak naprawd to nie jest potrzebne ;)
		Qinsert2(temp);
	}

	Tree=QextractMin();	
	// tutaj inicjujemy kod korzenia... na znaczek '\0'
	// a wlasciwie robi to za nas calloc
	Tree->code=(uint8_t*)calloc(1,sizeof(uint8_t));
	HuffmanMakeCodes(Tree);
}

// funkcja zlicza zawartosc pliku i umiesza w kolejce wszystkie literki z
// pliku a nastepnie wzywa HuffmanMakeTree 
void HuffmanInit(uint8_t *filename) {
	uint32_t *fileCharCount=NULL;
	uint32_t charsCount=0;

	fileCharCount=HuffmanCount(filename);
		
	for(int i=0;i<256;i++) 
		if(fileCharCount[i]!=0) charsCount++; 

	if(charsCount==0) {
		printf("Plik jest pusty!\n");
		exit(-1);
	}

	printf("%d unique chars found in file: %s\n",charsCount,filename);

	Qinit();	// inicjalizacja kolejki
		
	for(int i=0;i<256;i++)
		if(fileCharCount[i]!=0) 
			Qinsert(fileCharCount[i],(uint8_t)i);
			
	free(fileCharCount);

	HuffmanMakeTree(charsCount);	
}

int main(int argc, char *argv[]) {
	if(argc==2) HuffmanInit(argv[1]);
	else {
		printf("Usage: %s <nazwa_pliku>\n",argv[0]);
		return(-1);
	}

	printf("Real file lenght:\t\t %d bits\n",RealLen);
	printf("Compressed lenght:\t\t %d bits\n",HuffmanCompressedLen);
	printf("Compression ratio:\t\t %f\n",(float)HuffmanCompressedLen/(float)RealLen);

	return(0);
}
// (c) 2005 Witold Bolt
