// kompiluj z --std=c99 lub --std=gnu99

// wersja TN

// (c) 2005 Witold Bot
// www.hope.art.pl

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

// definicje typow
typedef struct {
	uint8_t c;
	uint32_t freq;
	uint32_t left;
	uint32_t right;
	uint8_t *code;
} TTrecord;


typedef struct Qrecord *Precord;
typedef struct Qrecord {
	uint32_t id;
	Precord next;
	Precord prev;
} TQrecord;

// globalne zminne dla drzewa i kolejki
TTrecord *tree=NULL;	// drzewo w postaci tabeli !
Precord QQ=NULL;

// operacje na kolejce
uint32_t QExtractMin() {
	uint32_t minI;
	Precord Queue=QQ;
	Precord Qmin=QQ;
	
	while(Queue != NULL) {
	if(tree[Queue->id].freq < tree[Qmin->id].freq) {
		Qmin=Queue;
		}
		Queue=Queue->next;
        }

        minI=Qmin->id;
	if(Qmin->prev!=NULL) Qmin->prev->next = Qmin->next;
	if(Qmin->next!=NULL) Qmin->next->prev = Qmin->prev;
	if(Qmin==QQ) QQ=Qmin->next;
	free(Qmin);
	return(minI);
}

void QInsert(uint32_t id) {
	Precord  new = (Precord)malloc(sizeof(TQrecord));
	new->next=QQ;
	if(QQ!=NULL)QQ->prev=new;
	new->prev=NULL;
	new->id = id;
	QQ=new;
}

// zliczanie wystapien znakow w pliku
uint32_t* freqCount(char *filename) {
	FILE *plik = fopen(filename,"rb");
	uint32_t *output=(uint32_t*)calloc(256,sizeof(uint32_t));
	uint8_t c;

	if (plik==NULL) { printf("Blad otwarcia pliku: %s\n",filename);
		exit(-1); }

	while((!feof(plik))&&(fread(&c,sizeof(uint8_t),1,plik))) {
	 output[(int)c]++;
        }

	fclose(plik);
	return(output);
}

// wyznaczanie kodow dla gotowego drzewa
void HuffmanCodesCalculate(uint32_t root) {
	uint32_t len = strlen(tree[root].code);

	tree[tree[root].left].code = 
	  (uint8_t*)calloc(len+2,sizeof(uint8_t));
	
	tree[tree[root].right].code = 
	  (uint8_t*)calloc(len+2,sizeof(uint8_t));

	strncpy(tree[tree[root].left].code, tree[root].code, len+1);
	tree[tree[root].left].code[len] = '0';

	strncpy(tree[tree[root].right].code, tree[root].code, len+1);
	tree[tree[root].right].code[len] = '1';

	if(tree[tree[root].left].left!=tree[tree[root].left].right) 
	HuffmanCodesCalculate(tree[root].left);

	if(tree[tree[root].right].left!=tree[tree[root].right].right) 
	HuffmanCodesCalculate(tree[root].right);
}

// glowna funkcja:
// wywyolje freqCount, tworzy drzewo i wywoluje generowanie kodow
uint32_t Huffman(char *filename) {
	uint32_t *freq=NULL;
	uint32_t charCount=0;
	uint32_t left,right,root;
	uint32_t j=0;

	freq=freqCount(filename);

	for(int i=0;i<256;i++) {if(freq[i]!=0) charCount++;}
	if(charCount==0) {printf("Plik pusty lub brak pliku!\n"); exit(-1);}
	printf("W pliku %s jest %d roznych znakow.\n",filename,charCount);

	tree = (TTrecord*)calloc(2*charCount,sizeof(TTrecord));

	for(int i=0;i<256;i++) 
	 	if(freq[i]!=0) {
				tree[j].c = (uint8_t)i;
				tree[j].freq=freq[i];
				tree[j].left=j;
				tree[j].right=j;
				QInsert(j);
				j++;
                              }

	free(freq);

	for(uint32_t i=0;i<charCount-1;i++) {
		left   = QExtractMin();
		right  = QExtractMin();
		tree[j].left=left;
		tree[j].right=right;
		tree[j].freq = tree[left].freq + tree[right].freq;
		QInsert(j);
		j++;
	}

	root = QExtractMin();
	
	tree[root].code = (uint8_t*)calloc(1,sizeof(uint8_t));
	HuffmanCodesCalculate(root);

	return(j);
}

int main(int argc, char *argv[]) {
        uint32_t j;
        uint32_t real_len=0;
        uint32_t comp_len=0;

	if(argc==2) j=Huffman(argv[1]);
	else {
		printf("podaj jeden argument (nazwa pliku)!\n");
		exit(-1);
	}	

	for(uint32_t i=0;i<j;i++) {
	 if((tree[i].left == tree[i].right)&&(tree[i].freq!=0)) {
#ifdef STATS
	 	printf("%c (%d):\t%d:\t%s (%d)\n",
	 	tree[i].c, tree[i].c, tree[i].freq, tree[i].code, strlen(tree[i].code));
#endif
	 comp_len+=tree[i].freq*strlen(tree[i].code);
	 real_len+=tree[i].freq*8;
	 }
	}
	printf("Rozmiar nieskompresowny:\t%d bits\n",real_len);
	printf("Rozmiar skompresowany:\t\t%d bits\n",comp_len);
	printf("Stopien kompresji:\t\t%f\n",(float)comp_len/(float)real_len);
	return(0);
}
// (c) 2005 Witold Bolt
