// (c) 2005 W. Bolt <ja@hope.art.pl>
// na bazie kodu napisanego przez dr P. Paczkowski <pmp@math.univ.gda.pl>


#include "tree.h"

Wskwezla nil;

/* funkcja inicjujaca nil; musi byc wywolana przed budowaniem drzewa */
void nilInit(void){
  nil= (Wskwezla) malloc(sizeof(Twezla));
  nil->ojciec = NULL;
  nil->kolor = BLACK;
  nil->lewy = nil->prawy = NULL;  
}

/* 2 do potegi n - fun. pomocnicza */
int potega(int n){
	if (n==0)  return 1;
	else       return 2*potega(n-1) ;
}

static Wskwezla buduj1(int n, int przesuniecie, Wskwezla ojciec){
    Wskwezla w = (Wskwezla) malloc(sizeof(Twezla));
    if (n==0)   w=nil;
    else {
    w->ojciec=ojciec;
    w->klucz = (potega(n-1)+przesuniecie)*1 - 1;
    w->kolor = (n%2 ? RED : BLACK);
    w->lewy = buduj1(n-1,przesuniecie,w);
    w->prawy = buduj1(n-1,przesuniecie+potega(n-1),w);
    }
    return w;
}

Wskwezla buduj(int n){
  Wskwezla w = buduj1(n+1,0,nil);
  w->kolor=BLACK;
  return w;
}

/* ----------------- szukanie elementu ---------------------- */

Wskwezla szukaj (int n, Wskwezla root) {
 if (root == NULL) return NULL;
 if (root->klucz == n ) return root;
 else if (root->klucz < n) return szukaj(n,root->prawy);
 else return szukaj(n,root->lewy);
}

/* ----------------- rotacje elementu -------------------- */

Wskwezla Left_Rotate(Wskwezla T,Wskwezla x)
{
 Wskwezla y;
 Wskwezla root;

 root=T;
 
 y=x->prawy;

 if (y!=nil) {
 x->prawy = y->lewy;
 (y->lewy)->ojciec = x;
 y->ojciec = x->ojciec;
 
 if (x->ojciec == nil) root=y;
 else if (x==x->ojciec->lewy) x->ojciec->lewy = y;
 else x->ojciec->prawy = y;
 y->lewy =x;
 x->ojciec = y; }
 else {
 printf("debug: niepoprawna rotacja: Left_Rotate(%d,%d)\n",T->klucz,x->klucz);
 }
 return root;
}


Wskwezla Right_Rotate(Wskwezla T,Wskwezla x)
{
 Wskwezla y;
 Wskwezla root;

 root=T;

 y=x->lewy;
 if(y!=nil) {
 x->lewy = y->prawy;
 y->prawy->ojciec = x;
 y->ojciec = x->ojciec;

 if (x->ojciec == nil) root=y;
 else if (x==x->ojciec->lewy) x->ojciec->lewy = y;
 else x->ojciec->prawy = y;

 y->prawy = x;
 x->ojciec = y; }
 else {
 printf("debug: niepoprawna rotacja: Right_Rotate(%d,%d)\n",T->klucz,x->klucz);
 }

 return root;
}

/* ----------------- dodawanie elementu --------------------- */

static Wskwezla InitNode(int key) {
	Wskwezla w;
	w = (Wskwezla)malloc(sizeof(Twezla));
	w->klucz = key;
	w->ojciec= nil;
	w->lewy = nil;
	w->prawy = nil;
	return w;
}

static Wskwezla RB_Insert_Fixup(Wskwezla T,Wskwezla z) {
	Wskwezla root,y;

	root=T;

	while (z->ojciec->kolor == RED) {
		if (z->ojciec == z->ojciec->ojciec->lewy) {
			y = z->ojciec->ojciec->prawy;
			if (y->kolor == RED) {
				 z->ojciec->kolor = BLACK;
				 y->kolor = BLACK;
				 z->ojciec->ojciec->kolor = RED;
				 z = z->ojciec->ojciec;
			}
			else {if (z == z->ojciec->prawy) {
				z = z->ojciec;
				root=Left_Rotate(root,z); }
				z->ojciec->kolor = BLACK;
				z->ojciec->ojciec->kolor = RED;
				root=Right_Rotate(root,z->ojciec->ojciec);
			     }
		}
		else {
			y = z->ojciec->ojciec->lewy;
			if (y->kolor == RED) {
				 z->ojciec->kolor = BLACK;
				 y->kolor = BLACK;
				 z->ojciec->ojciec->kolor = RED;
				 z= z->ojciec->ojciec;
			}
			else {if (z == z->ojciec->lewy) {
				z = z->ojciec;
				root=Right_Rotate(root,z); }
				z->ojciec->kolor = BLACK;
				z->ojciec->ojciec->kolor = RED;
				root=Left_Rotate(root,z->ojciec->ojciec);
			     }
		}
	}
	root->kolor = BLACK;
	return(root);
}


Wskwezla RB_Insert(Wskwezla T,int key) {
	Wskwezla root,x,y,z;

	root=T;
	z = InitNode(key);

	y=nil;
	x=root;
	while (x!=nil) {
		y = x;
		if(z->klucz < x->klucz) x=x->lewy;
			else x=x->prawy;
		}
	z->ojciec = y;

	if(y==nil)
		root=z;
	else if (z->klucz < y->klucz)
		y->lewy = z;
	     else
	     	y->prawy = z;
	z->lewy = nil;
	z->prawy = nil;
	z->kolor = RED;	

	root = RB_Insert_Fixup(root,z);
	return root;
}
/* ----------------- wysokosc drzewa ------------------------ */

int TreeHeight(Wskwezla w)
{
  int lewa,prawa;

  if (w==nil) return(-1);

  lewa = TreeHeight(w->lewy);
  prawa = TreeHeight(w->prawy);

  if(lewa>prawa) return(lewa+1);
  else return(prawa+1);
}

/* ----------------- najwiekszy klucz w drzewie -------------- */

int TreeMax(Wskwezla w)
{
 if (w->prawy==nil) return(w->klucz);
 else return(TreeMax(w->prawy));
}

/* ----------------- najmniejszy klucz w drzewie -------------- */

int TreeMin(Wskwezla w)
{
 if (w->lewy==nil) return(w->klucz);
 else return(TreeMax(w->lewy));
}
